# -*- coding: utf-8 -*-

from odoo import api, models, fields, _
from odoo.exceptions import UserError

from collections import defaultdict
from markupsafe import Markup

import logging
_logger = logging.getLogger(__name__)

REASON_MESSAGE_HEADER = _("Plan switch reason(s):")
CUSTOM_PLAN_REASON_MESSAGE = {
    'hosting': _("the database is hosted on Odoo.sh or on premise"),
    'studio': _("Studio has been installed"),
    'multi_company': _("a second company has been detected"),
    #'rpc': _("Rpc call to the database have been detected"),
    'customization': _("Customization (lines of code) has been detected on the database"),
    'database_unknown': _("The database is not linked to the subscription, we consider then the hosting of the database is on premise"),
    'hosting_unkown': _("The hosting is unknown, we consider then the hosting of the database is on premise"),
    'base_automation': _("Automation tools has been installed"),
}


class SaleOrderTemplate(models.Model):
    _inherit = 'sale.order.template'

    custom_plan_id = fields.Many2one('sale.order.template')


class SaleOrder(models.Model):
    _inherit = 'sale.order'

    @api.depends('order_line.invoice_lines')
    def _get_invoiced(self):
        AccountMove = self.env['account.move']

        other_invoices = AccountMove.search([
            ('line_ids.subscription_id', 'in', self.ids),
            ('move_type', 'in', ('out_invoice', 'out_refund')),
        ])
        inv_by_subid = defaultdict(AccountMove.browse)
        for invoice in other_invoices:
            for s in invoice.line_ids.mapped("subscription_id"):
                inv_by_subid[s.id] |= invoice

        super()._get_invoiced()

        for order in self:
            invoices = order.invoice_ids | inv_by_subid[order.id]
            order.invoice_ids = invoices
            order.invoice_count = len(invoices)

    def _confirm_subscription(self):
        super()._confirm_subscription()
        self.check_user_discount()

    def check_user_discount(self):
        user_products = self.env.ref('openerp_enterprise.product_odoo_user_standard') + self.env.ref('openerp_enterprise.product_odoo_user_custom')
        discounted_user_products = self.env.ref('openerp_enterprise.product_odoo_user_standard_discount') + self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')

        for sub in self:
            discounts = sum(sub.order_line.filtered(lambda l: l.product_id in discounted_user_products).mapped('price_subtotal'))
            users = sum(sub.order_line.filtered(lambda l: l.product_id in user_products).mapped('price_subtotal'))
            if -discounts > users * 0.25:
                raise UserError("Discount User cannot exceed 25%")


    ###   UPSELL   ###
    def _confirm_upsell(self):
        self = self.with_context(skip_upsell_plan_check=True, no_check=True)
        super()._confirm_upsell()
        standard_products = self.env.ref('openerp_enterprise.product_odoo_user_standard') + self.env.ref('openerp_enterprise.product_odoo_user_standard_discount')
        custom_products = self.env.ref('openerp_enterprise.product_odoo_user_custom') + self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
        for so in self:
            if so.subscription_id.sale_order_template_id.custom_plan_id \
                    and (so.subscription_id.order_line.mapped('product_id') & custom_products):  # if standard with line custo

                for ln in so.subscription_id.order_line.filtered(lambda line: line.product_id in standard_products):
                    ln.unlink()
                so.subscription_id.sale_order_template_id = so.subscription_id.sale_order_template_id.custom_plan_id.id
            so.subscription_id._merge_discount()
            if not self.env.su:
                so.subscription_id._check_sub()

    def inverse_product_user(self, product):
        return {
            self.env.ref('openerp_enterprise.product_odoo_user_standard'): self.env.ref('openerp_enterprise.product_odoo_user_custom'),
            self.env.ref('openerp_enterprise.product_odoo_user_standard_discount'): self.env.ref('openerp_enterprise.product_odoo_user_custom_discount'),
            self.env.ref('openerp_enterprise.product_odoo_user_custom'): self.env.ref('openerp_enterprise.product_odoo_user_standard'),
            self.env.ref('openerp_enterprise.product_odoo_user_custom_discount'): self.env.ref('openerp_enterprise.product_odoo_user_standard_discount'),
        }.get(product)

    def action_upsell_switch_plan(self):
        for l in self.order_line.filtered(
            lambda line: line.product_template_id.categ_id == line.env.ref('openerp_enterprise.product_category_odoo_users')
            and not line.product_template_id.is_odoo_app
        ):
            product_id = self.inverse_product_user(l.product_id)
            if product_id:
                self.env['sale.order.line'].create({
                    'order_id': self.id,
                    'product_id': product_id.id,
                    'product_uom_qty': l.parent_line_id.product_uom_qty + l.product_uom_qty,
                    'discount': l.discount,
                })
                l.product_uom_qty = l.parent_line_id.product_uom_qty * -1
        if not self._get_maintenance_line():
            self.action_activate_maintenance()


    ###   PLAN   ###
    def _get_plan_reason(self):
        if self.hosting in ('paas', 'premise'):
            return ['hosting']
        if self.order_line.filtered(lambda l: l.product_template_id.categ_id.id == 57):  # Odoo.sh
            return ['hosting']
        if self.hosting == 'saas':
            if len(self.database_ids) == 1:
                return self.database_ids[0]._get_plan_reason()
            return ['database_unknown']
        return ['hosting_unkown']  # in doubt

    def get_plan(self):
        return self._get_plan_reason() and 'custom' or 'standard'

    def _plan2022_need_upsell(self):
        self.ensure_one()
        if self.plan_id.billing_period_unit != 'year':
            return False
        if self._maintenance_need_upsell():
            return True
        if self.need_odoo_sh_upsell():
            return True
        if self.database_ids[:1].quantity > self._get_billed_user_count():
            return True
        if self.database_ids[:1].iot_box_count > self._get_billed_iot_box_count():
            return True
        if self._is_standard_plan() and self.get_plan() == 'custom':
            return True

        return False

    def _get_custom_plan_message(self):
        reasons = '\n'.join([f"\t{CUSTOM_PLAN_REASON_MESSAGE[r]}" for r in self._get_plan_reason()])
        return f"{REASON_MESSAGE_HEADER}\n{reasons}"

    def _get_custom_plan_message_html(self):
        reasons = '\n'.join([f"<li>{CUSTOM_PLAN_REASON_MESSAGE[r]}</li>" for r in self._get_plan_reason()])
        return f"""<ul>
            <li>{REASON_MESSAGE_HEADER}
                <ul>{reasons}</ul>
            </li>
        </ul>"""

    def _get_product_diff(self, product, actual_quantity):
        current_lines = self.order_line.filtered(lambda l: l.product_id == product)
        upsell_lines = self.upsell_order_id.order_line.filtered(lambda l: l.product_id == product)
        diff = actual_quantity - sum(l.product_uom_qty for l in upsell_lines | current_lines)
        return int(diff)

    def _get_invoice_diff(self):
        """
        Analyze subscriptions and their upsell SO to determine what upsell is needed.

        Parse a subscription and its upsell SO lines and compare it to the current
        state of the database.

        :return: dict where the value for each id is a dict containing:
            database_id: the database_id the contract was compared to
            upsell: a list of values dict containing the necessary info
                    to create the upsell lines (pluggable in a command call)
            or
            False if no upsell is needed for that subscription
        :rtype: dict
        """
        res = dict.fromkeys(self.ids, False)

        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')
        for subscription in self.filtered(
                lambda s: s._is_enterprise_plans() and
                s.plan_id.billing_period_unit == 'year'
        ):
            user_product = self.env.ref('openerp_enterprise.product_odoo_user_custom')
            if subscription._is_standard_plan():
                user_product = self.env.ref('openerp_enterprise.product_odoo_user_standard')

            db = self.database_ids[:1]
            user_diff = subscription._get_product_diff(user_product, db.quantity)
            upsell_line = []
            if user_diff > 0:
                upsell_line.append({
                    'product_uom_qty': user_diff,
                    'product_id': user_product.id,
                    'product_uom': subscription.sale_order_template_id.get_product_uom(user_product).id,
                    'name': user_product.name,
                })

            iot_diff = subscription._get_product_diff(iot_box_product, db.iot_box_count)
            if iot_diff > 0:
                upsell_line.append({
                    'product_uom_qty': iot_diff,
                    'product_id': iot_box_product.id,
                    'product_uom': subscription.sale_order_template_id.get_product_uom(iot_box_product).id,
                    'name': iot_box_product.name,
                })

            need_to_switch_plan = False
            if self._is_standard_plan() and self.get_plan() == 'custom':
                need_to_switch_plan = True

            upsell_line.extend(subscription._get_odoo_sh_invoice_diff())

            if upsell_line or need_to_switch_plan:
                res[subscription.id] = {
                    'database_id': db.id,
                    'upsell': upsell_line,
                }

            if need_to_switch_plan:
                res.get(subscription.id)['switch_plan'] = True
                res.get(subscription.id)['upsell'].append({
                    'display_type': 'line_note',
                    'sequence': 9999,
                    'name': subscription._get_custom_plan_message(),
                })

        return res

    def set_apps_recurring_lines(self):
        def switch_plan(subscription):
            user_category = self.env.ref('openerp_enterprise.product_category_odoo_users')
            subscription.sale_order_template_id = subscription.sale_order_template_id.custom_plan_id.id
            for line in subscription.order_line.filtered(lambda l: l.product_template_id.categ_id == user_category):
                self.env['sale.order.line'].create({
                    'order_id': subscription.id,
                    'product_id': subscription.inverse_product_user(line.product_id).id,
                    'product_uom_qty': line.product_uom_qty,  # Correct qty are updated later
                    'discount': line.discount,
                    'computation': line.computation,
                })
                line.unlink()
            if not subscription._get_maintenance_line():
                subscription.action_activate_maintenance()
            subscription.message_post(
                body=Markup(subscription._get_custom_plan_message_html()),
                message_type='comment',
                subtype_xmlid='mail.mt_comment',
            )

        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')
        monthly_plans = self.filtered(
            lambda s: s._is_enterprise_plans() and s.plan_id.billing_period_unit == 'month')
        for subscription in monthly_plans:
            sh_values = subscription._prepare_apps_product_odoo_sh()
            if sh_values:  # Avoiding useless write
                subscription.write({'order_line': sh_values})

            if not subscription.database_ids:
                continue

            if subscription._is_standard_plan() and subscription.get_plan() == 'custom':
                switch_plan(subscription.with_context(skip_upsell_plan_check=True))
                subscription._check_sub()

            db = subscription.database_ids[:1]
            user_product = self.env.ref('openerp_enterprise.product_odoo_user_custom')
            discount_product = self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
            if subscription._is_standard_plan():
                user_product = self.env.ref('openerp_enterprise.product_odoo_user_standard')
                discount_product = self.env.ref('openerp_enterprise.product_odoo_user_standard_discount')

            user_line = subscription.order_line.filtered(lambda l: l.product_id == user_product)[:1]
            discount_line = subscription.order_line.filtered(lambda l: l.product_id == discount_product)[:1]
            user_sold = user_line.product_uom_qty
            user_db = db.quantity
            if (user_sold != user_db and user_line.computation != 'manual') or (user_db > user_sold):
                user_line.product_uom_qty = user_db

            # In case of user downsell, downsell the first year discount as well
            if user_sold > user_db and user_line.computation != 'manual' and discount_line:
                discount_line.product_uom_qty = user_db

            if db.iot_box_count:
                iot_lines = subscription.order_line.filtered(lambda l: l.product_id == iot_box_product)[:1]
                iot_diff = int(db.iot_box_count - iot_lines.product_uom_qty)
                if iot_lines and ((iot_diff and iot_lines.computation != 'manual') or iot_diff):
                    iot_lines.product_uom_qty = db.iot_box_count
                elif iot_diff:
                    self.env['sale.order.line'].create({
                        'order_id': subscription.id,
                        'product_id': iot_box_product.id,
                        'product_uom_qty': db.iot_box_count,
                        'computation': 'automatic',
                    })

        return True

    def _prepare_renew_upsell_order(self, subscription_state, message_body):
        # need to get the invoice diff before the creation of the upsell_order otherwise result will be wrong
        res = self._get_invoice_diff()
        action = super()._prepare_renew_upsell_order(subscription_state, message_body)

        if subscription_state == '7_upsell':
            if res.get(self.id) and res[self.id].get('switch_plan'):
                sale_order = self.browse(action['res_id'])
                sale_order.action_upsell_switch_plan()
        return action

    def _merge_discount(self):
        self.ensure_one()
        discount_products = (
            self.env.ref('openerp_enterprise.product_official_partner_discount') |
            self.env.ref('openerp_enterprise.product_learning_partner_discount') |
            self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount') |
            self.env.ref('openerp_enterprise.product_multiyear_discount')
        )
        for discount_product in discount_products:
            discount_lines = self.order_line.filtered(
                lambda l, p=discount_product: l.product_uom_qty == 1 and l.product_id == p
            )
            if len(discount_lines) < 2:
                continue

            discount_lines[0].price_unit = sum(discount_lines.mapped('price_unit'))
            discount_lines[1:].unlink()


class SaleOrderPricingConvert(models.Model):
    _inherit = 'sale.order'

    old_sub_id = fields.Many2one('sale.order')

    ###   CONVERT   ###
    def get_lines_distributed(self):
        user_products = self.env.ref('openerp_enterprise.product_odoo_user_standard') + self.env.ref('openerp_enterprise.product_odoo_user_custom')
        distributed = defaultdict(self.env['sale.order.line'].browse)
        new_amount = 0

        for l in self.order_line.filtered('product_id'):
            if l.product_template_id.id == 4929 and l.price_subtotal == 0:  # mx have initial user discount = 0
                distributed['init_discount'] += l

            elif l.price_subtotal < 0:
                if l.product_template_id.id in (10241, 10240, 4926):  # Subscription Discount (official/learning partner) | Odoo.sh Access
                    distributed['discount_recu'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 4929:  # Discount User online
                    distributed['init_discount'] += l
                    new_amount += l.price_subtotal
                elif l.product_id.categ_id.id == 42:  # Old Discount apps
                    distributed['apps'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id in (16523, 16523):  # DISCOUNT misc + POS DISCOUNT?
                    distributed['once_discount'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id in (20876,):  # Mnt adjustement
                    distributed['mnt_adjust'] += l
                    new_amount += l.price_subtotal
                else:
                    distributed['WTF_discount'] += l
                    new_amount += l.price_subtotal
                    raise UserError('Unknown line discount...')
            else:
                if l.product_template_id.id == 8416:  # TODO BEFORE CATEG APPS 42 -> IOT SUBSCRIPTION
                    distributed['misc_recu'] += l
                    new_amount += l.price_subtotal
                elif l.product_id.categ_id.id == 42:  # apps...
                    distributed['apps'] += l
                    new_amount += l.price_subtotal
                    if l.product_template_id.id == 2268 and not self.env.context.get('no_maintenance_warning'):  # Maintenance apps (used wrongly)
                        if l.product_uom_qty > 1:
                            raise UserError('Wrong maintenance configuration - Qty > 1')
                        elif distributed['users'] and l.price_subtotal > 1.25 * max(distributed['users'].mapped('price_unit')):
                            raise UserError('Wrong maintenance configuration - Cost not normal - Probably the wrong product has been used.')
                elif l.product_template_id.id in (4928, 199, 783, 205, 211):  # (User online, Odoo Enterprise User, Odoo Online user, Odoo Enterprise Yearly, Odoo Enterprise Yearly (renewal))
                    distributed['users'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id in (4929,):  # discount user wrongly used for users
                    distributed['users'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 408:
                    distributed['xtra_company'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id in (11690, 20875):  # Maintenance_by_loc
                    distributed['mnt_loc'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 20876:  # Maintenance Adjustement
                    distributed['mnt_adjust'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 48 and (l.product_template_id.name.startswith('Maintenance of') or l.product_template_id.name.startswith('Customized ')):
                    #distributed['mnt_custom_of'] += l
                    distributed['mnt_adjust'] += l
                    new_amount += l.price_subtotal

                elif l.product_template_id.id == 10745:  # Custom contract clause coverage
                    distributed['misc_recu'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 386:  # [Online] Specific Development # RECU
                    distributed['misc_recu'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 57:
                    distributed['odoosh'] += l
                    new_amount += l.price_subtotal

                elif l.product_template_id.categ_id.id == 2:  # certif
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 4:  # Misc
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 811:  # Extra Fee
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.id == 8937:  # POS
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 51:  # IAP
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 40:  # Odoo exp - tour
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 27:  # PS service (mise en prod, ...)
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 45:  # appstore
                    distributed['misc_once'] += l
                    new_amount += l.price_subtotal

                elif l.product_template_id.categ_id.id == 47:  # spack
                    distributed['spack'] += l
                    new_amount += l.price_subtotal
                elif l.product_template_id.categ_id.id == 17:  # Partnership
                    distributed['misc_recu'] += l
                    new_amount += l.price_subtotal
                    #raise UserError('Partnership...')
                else:
                    distributed['WTF_NON_DISCOUNT'] += l
                    if l.product_template_id.id in user_products.ids:
                        raise UserError('Already migrated')
                    if l.price_subtotal != 0:
                        raise UserError('Unknown line... %s in %s' % (l.product_id.name, self.name))

        if abs(new_amount - self.amount_untaxed) > 0.02:
            raise UserError("wrong distribution ...")
        return distributed

    def list_template(self):
        return {
            34: self.env.ref('openerp_enterprise.standard_template_odoo_monthly').id,   # 34 = openerp_enterprise.quotation_template_odoo_monthly
            35: self.env.ref('openerp_enterprise.standard_template_odoo_yearly').id,   # 35 = openerp_enterprise.quotation_template_odoo_yearly
            36: self.env.ref('openerp_enterprise.standard_template_odoo_2_years').id,   # 36 = openerp_enterprise.quotation_template_odoo_2_years
            37: self.env.ref('openerp_enterprise.standard_template_odoo_3_years').id,   # 37 = openerp_enterprise.quotation_template_odoo_3_years
            38: self.env.ref('openerp_enterprise.standard_template_odoo_4_years').id,   # 38 = openerp_enterprise.quotation_template_odoo_4_years
            33: self.env.ref('openerp_enterprise.standard_template_odoo_5_years').id,   # 33 = openerp_enterprise.quotation_template_odoo_5_years
            58: self.env.ref('openerp_enterprise.standard_template_odoo_monthly').id,   # 58 = openerp_enterprise.contract_template_odoo_online

            59: self.env.ref('openerp_enterprise.standard_template_odoo_monthly').id,
            61: self.env.ref('openerp_enterprise.standard_template_odoo_yearly').id,
            63: self.env.ref('openerp_enterprise.standard_template_odoo_yearly').id,
            65: self.env.ref('openerp_enterprise.standard_template_odoo_yearly').id,
            67: self.env.ref('openerp_enterprise.standard_template_odoo_3_years').id,
            68: self.env.ref('openerp_enterprise.standard_template_odoo_3_years').id,
            71: self.env.ref('openerp_enterprise.standard_template_odoo_5_years').id,
            73: self.env.ref('openerp_enterprise.standard_template_odoo_2_years').id,
            76: self.env.ref('openerp_enterprise.standard_template_odoo_2_years').id,
            77: self.env.ref('openerp_enterprise.standard_template_odoo_4_years').id,
        }

    def get_template(self, old, plan):
        old_id_to_new_record = self.list_template()

        tpl_id = old_id_to_new_record.get(old)

        if not tpl_id and self.subscription_id:
            # if migrated a child, check the parent sub
            tpl_id = self.subscription_id.sale_order_template_id.id
            plan = "standard"  # bypass the search of matching custom plan

        if not tpl_id:
            tpl_id = old_id_to_new_record[35]
            raise UserError("wrong tmpl %s" % old)

        if self.env['sale.order.template'].browse(tpl_id).plan_id != self.plan_id and not self.env.context.get('no_check_recurrence'):
            raise UserError('Wrong recurrence matching between plan and sub.')

        tpl = self.env['sale.order.template'].browse(tpl_id)
        if plan == "custom":
            if not tpl.custom_plan_id:
                raise UserError('Wrong plan matching')
            tpl = tpl.custom_plan_id
        return tpl

    def action_convert_all_to_new_pricing(self):
        if not set(self.order_line.product_template_id.ids) & set([4928, 199, 783, 205, 211]):
            raise UserError('No way to convert it, users missing')

        self._convert_new_pricing()
        plan = self._is_standard_plan() and "standard" or "custom"
        for sub in self.subscription_child_ids:
            if sub.state in ('draft', 'sent') and not sub._is_enterprise_plans():
                sub._convert_new_pricing(force_plan=plan)
                if sub.subscription_state == '7_upsell' and sub.amount_untaxed == 0:
                    sub._action_cancel()

    def _convert_new_pricing(self, force_plan=None):
        SOL = self.env['sale.order.line']
        louch = ''
        lost_amount = 0
        plan = force_plan or self.get_plan()
        prices = {l['id']: l['price_unit'] for l in self.order_line.read(['price_unit'])}

        # duplicate, tag, invoice_lines and cancel the duplicate
        self.old_sub_id = self.copy(dict(
            website_id=False,
            tag_ids=[(4, self.env.ref('openerp_enterprise.old_sub_bkp').id, 0)],
        ))
        self.old_sub_id.action_cancel()

        # get details
        splitted = self.get_lines_distributed()

        lines_to_remove = self.order_line.filtered(lambda p: not p.product_id or p.product_uom_qty == 0)
        lines_to_remove -= splitted['mnt_loc']

        olduserprice = sum((splitted['users'] + splitted['init_discount']).mapped('price_subtotal'))

        lost_app = 0
        for l in (splitted['apps'] + splitted['xtra_company']):
            lines_to_remove += l
            lost_app += l.price_subtotal
        lost_amount += lost_app

        # FROM HERE, you cannot base anymore on old price from sol, use prices dict instead
        new_template = self.get_template(self.sale_order_template_id.id, plan)
        self.write({
            'sale_order_template_id': new_template,
            'plan_id': new_template.plan_id,
        })

        # USER
        user_lines = SOL
        if splitted['users']:
            SOL.create({
                'order_id': self.id,
                'product_id': False,
                'display_type': "line_section",
                'name': "Licence - Users",
                'sequence': 10,
            })
            for l in splitted['users']:
                lines_to_remove += l
            new_line = SOL.create({
                'order_id': self.id,
                'product_id': self.env.ref('openerp_enterprise.product_odoo_user_%s' % plan).id,
                'product_uom_qty': sum(splitted['users'].mapped('product_uom_qty')),
                'discount': min(splitted['users'].mapped('discount')),
                'sequence': 11,
            })
            user_lines += new_line  # need to keep to recompute user with discounted price in rare condition like monthly increased

            lines_to_remove += splitted['init_discount']

            lost_user = olduserprice - sum(user_lines.mapped('price_subtotal'))
            lost_amount += lost_user

        # ODOO.SH
        if splitted['odoosh']:
            SOL.create({
                'order_id': self.id,
                'product_id': False,
                'display_type': "line_section",
                'name': "Hosting: Odoo.sh",
                'sequence': 20,
            })
            for l in splitted['odoosh']:
                l.write({
                    'price_unit': prices[l.id],
                    'sequence': 21,
                })

        # RECURRENT
        if splitted['misc_recu'] + splitted['discount_recu']:
            SOL.create({
                'order_id': self.id,
                'product_id': False,
                'display_type': "line_section",
                'name': "Recurrent",
                'sequence': 30,
            })
            for l in (splitted['misc_recu'] + splitted['discount_recu']):
                l.write({
                    'price_unit': prices[l.id],
                    'sequence': 35,
                })

        # MAINTENANCE
        if splitted['mnt_adjust'] or splitted['mnt_loc']:
            if (splitted['mnt_adjust'] + splitted['mnt_loc']).filtered(lambda l: plan == 'standard' and 'product_uom_qty' or True):  # copy loc of 0 only in custom plan
                SOL.create({
                    'order_id': self.id,
                    'product_id': False,
                    'display_type': "line_section",
                    'name': "Maintenance",
                    'sequence': 50,
                })
                for l in splitted['mnt_loc']:
                    l.write({
                        'sequence': 51,
                        'price_unit': prices[l.id],
                    })

                if splitted['mnt_adjust']:
                    splitted['mnt_adjust'][0].write({
                        'product_id': 18306,  # Maintenance Adjustment  # product.template 20876
                        'price_unit': sum([prices[l.id] for l in splitted['mnt_adjust']]),
                        'sequence': 52,
                    })
                    lines_to_remove += splitted['mnt_adjust'][1:]

        # OLD DISCOUNT ONCE - REMOVED FROM SO
        if splitted['once_discount']:
            # SOL.create({
            #     'order_id': self.id,
            #     'product_id': False,
            #     'display_type': "line_section",
            #     'name': "Old discount not recurring (LOST) - %s" % sum(splitted['once_discount'].mapped('price_subtotal')),
            #     'sequence': 60,
            # })
            lines_to_remove += splitted['once_discount']

        # MISC NON RECURRING LINES
        if splitted['spack'] + splitted['appstore'] + splitted['misc_once'] + splitted['ps']:
            SOL.create({
                'order_id': self.id,
                'product_id': False,
                'display_type': "line_section",
                'name': "Others",
                'price_unit': 0,
                'sequence': 65,
            })
            for l in (splitted['spack'] + splitted['appstore'] + splitted['misc_once'] + splitted['ps']):
                l.write({
                    'price_unit': prices[l.id],
                    'sequence': 66,
                })

        if splitted['WTF_discount']:
            for l in splitted['WTF_discount']:
                l.name = "*#*#*##*#*# %s " % l.name

            louch = "strange discount"
            raise UserError(louch)

        if lost_amount < 0:
            if self.plan_id.billing_period_unit != 'month':
                SOL.create({
                    'order_id': self.id,
                    'product_id': False,
                    'display_type': "line_section",
                    'name': "Subscription Adjustment",
                    'sequence': 70,
                })
                SOL.create({
                    'order_id': self.id,
                    'product_id': 36555,  # subscription_discount_recurrent - Subscription extra discount

                    'product_uom': self.env.ref('uom.product_uom_unit').id,
                    'price_unit': lost_amount,
                    'sequence': 71,
                })

        # CLEAN OLD OUTDATED LINE
        lines_to_remove.unlink()

        if self.amount_untaxed < 0 and splitted['discount_recu']:
            discount_to_remove = self.amount_untaxed
            for l in self.order_line.filtered(lambda l: l.product_template_id.id in (10240, 10241)):
                # import ipdb; ipdb.set_trace()
                if discount_to_remove:  # negative amount
                    old_pu = l.price_unit
                    l.price_unit = min(0, l.price_unit - discount_to_remove)
                    discount_to_remove -= (old_pu - l.price_unit)
                else:
                    break
            if discount_to_remove:
                raise UserError("Discounted lines to be fixed manually by saleman")



        return self.sale_order_template_id, self.pricelist_id

class OpenerpEnterpriseDatabase(models.Model):

    _inherit = 'openerp.enterprise.database'

    def get_plan(self):
        self.ensure_one()
        return self._get_plan_reason() and 'custom' or 'standard'

    def _module_installed(self, module_name):
        self.ensure_one()
        return self.app_ids.filtered(
            lambda app: app.module_id.technical_name == module_name and app.installed
        )

    def _get_plan_reason(self):
        self.ensure_one()
        reason = []
        if self.hosting in ('paas', 'premise'):
            reason.append('hosting')
        if self._module_installed('web_studio'):
            reason.append("studio")
        if self.info and '"company_count":' in self.info and 'company_count": 1,' not in self.info:
            reason.append("multi_company")
        # if '"rpc_call": true' in self.info:
        #     reason.append('rpc')
        if self.maintenance_lines_count:
            reason.append('customization')
        if not self._module_installed('sale_subscription') and self.has_base_automation:
            reason.append("base_automation")
        return reason
